import os

from gpubs.log import msg2


def extract_data(element):
    #import xml.etree.ElementTree as ET
    import lxml.etree as ET

    data = {}
    data["PMID"] = element.findtext("MedlineCitation/PMID")
    data["Title"] = element.findtext("MedlineCitation/Article/ArticleTitle")
    
    # get all the AbstractText's
    def get_abstract(abstract_elements):

        abstract_texts = []

        for abstract_element in abstract_elements:
            label = abstract_element.get('Label')
            nlm_category = abstract_element.get('NlmCategory')

            if label:
                label_text = f"[{label}] "
            elif nlm_category:
                label_text = f"[{nlm_category}] "
            else:
                label_text = ""

            # Serialize the element to a string and decode it to avoid truncating text
            element_text = ET.tostring(abstract_element, method='text', encoding='unicode')

            # Replace tab characters with 4 spaces directly on the element's text content
            element_text = element_text.replace('\t', '    ').strip()

            # Concatenate the label text and element text
            abstract_texts.append(f"{label_text}{element_text}")

        # Join all the abstract texts together
        abstract_text = " ".join(abstract_texts)
        return abstract_text

    data["Abstract"] = get_abstract(element.findall("MedlineCitation/Article/Abstract/AbstractText"))
    
    data["Journal"] = element.findtext("MedlineCitation/Article/Journal/Title")
    data["PublicationDate"] = element.findtext(
        "MedlineCitation/Article/Journal/JournalIssue/PubDate/Year"
    )
    data["JournalTitle"] = element.findtext("MedlineCitation/Article/Journal/Title")
    data["ArticleType"] = element.findtext(
        "MedlineCitation/Article/PublicationTypeList/PublicationType"
    )

    # Extract the descriptor names and qualifier names from the XML
    mesh_headings = element.findall(".//MeshHeading")
    mesh_heading_list = []
    for heading in mesh_headings:
        descriptor_name = heading.findtext("DescriptorName")
        qualifier_names = [
            qualifier.text for qualifier in heading.findall("QualifierName")
        ]
        mesh_heading_list.append(descriptor_name)
        mesh_heading_list.extend(qualifier_names)
    data["MeshHeadingList"] = ",".join(mesh_heading_list)

    publication_types = element.findall(
        "MedlineCitation/Article/PublicationTypeList/PublicationType"
    )
    data["PublicationTypeList"] = ",".join([ptype.text for ptype in publication_types])

    # find primary authors (1st, last, 2nd, next to last)
    authors_list = element.findall("MedlineCitation/Article/AuthorList/Author")
    # Sort authors based on their order within the list
    authors = [
        author.findtext(".//LastName", default="")
        + ","
        + author.findtext(".//ForeName", default="")
        + " "
        + author.findtext(".//Initials", default="")
        for author in sorted(authors_list, key=lambda x: authors_list.index(x))
    ]

    data["Authors"] = ";".join(authors)
    # data['FirstAuthor'] = authors[1] if authors else None  # Get the first author if authors list is not empty
    # data['SecondAuthor'] = authors[2] if authors else None  # Get second author if authors list is not empty
    # data['NextLastAuthor'] = authors[-2] if authors else None  # Get the second to last author if authors list is not empty
    data["LastAuthor"] = (
        authors[-1] if authors else None
    )  # Get the last author if authors list is not empty

    return data


def prune_df(df, length_threshold=405, verbose=2):
    # exclude articles with no abstract, no date, or abstracts that are too short (less than length_threshold letters)
    pruned_df = df[df["Abstract"].notna() & df["PublicationDate"].notna()]

    # cut out any short articles
    all_pruned = len(pruned_df)
    msg2(
        verbose, f"Number of all abstracts before pruning short articles = {all_pruned}"
    )
    pruned_df = pruned_df[pruned_df["Abstract"].str.len() >= length_threshold]
    long_pruned = len(pruned_df)
    msg2(verbose, f"Number after pruning short articles = {long_pruned}")
    msg2(verbose, f"Number discarded for being too short: {all_pruned - long_pruned}")

    return pruned_df


def get_pub_df(filename, inpath, outpath, length_threshold, prune=True, verbose=0):
    import gzip
    # import xml.etree.ElementTree as ET
    import lxml.etree as ET
    import pandas as pd

    pubmed_filepath = os.path.join(inpath, filename)
    # Open the gzip'd XML file
    with gzip.open(pubmed_filepath, "rb") as f:
        # Read the contents of the gzip'd file
        gzip_content = f.read()

    # Parse the XML content using ElementTree
    root = ET.fromstring(gzip_content)

    # Extract data from each article and store in a list
    articles = []
    for article in root.findall(".//PubmedArticle"):
        articles.append(extract_data(article))

    # Create a DataFrame from the list of articles
    df = pd.DataFrame(articles)
    df = df.drop_duplicates()
    if prune:
        msg2(verbose, f"Number of all articles:{len(df)}")
        df = prune_df(df, length_threshold=length_threshold, verbose=verbose)
        msg2(verbose, f"Number of pruned articles:{len(df)}")

    # convert objects to simple types
    df["PublicationDate"] = df["PublicationDate"].astype(int)

    return df


def parse_pubs(inpath, outpath, length_threshold, csv_list, verbose=0):
    import glob

    for filepath in glob.glob(os.path.join(inpath, "pubmed*.xml.gz")):
        msg2(verbose, f"Converting file {filepath}")
        if os.path.isfile(filepath):
            filename = os.path.basename(filepath)
            df = get_pub_df(
                filename=filename,
                inpath=inpath,
                outpath=outpath,
                prune=True,
                length_threshold=length_threshold,
                verbose=verbose,
            )
            csv_filepath = os.path.join(outpath, f"{filename}.csv")
            df.to_csv(csv_filepath, header=False, index=False, sep="\t")
            msg2(verbose, f"Wrote file:{csv_filepath}")
            csv_list.append(csv_filepath)

    return csv_list



from typing import List
import os

from gpubs.models import ReferenceData
from gpubs.log import msg1, msg2
from gpubs.reference import download_gene_symbols, extract_gene_data
from gpubs.search_words import (
    fetch_brown_corpus,
    create_stop_words,
    read_search_stop,
    filter_search_terms,
)
from gpubs.fetch import check_disk_space, download_file, verify_md5
from gpubs.parse import parse_pubs


def create_gene_reference_data(m: ReferenceData):
    """Gets gene reference data into a convenient format for creating a m.filtered_terms_files, and ultimately, tagging abstracts with gene information.

    Performs the following:

      1. Downloads gene_info.gz from m.ncbi_gene_info_url

      2. Parses gene_info.gz to create the following files:

        - m.reference_path()/m.gene_symbols_filename [default = v1/reference/gene_symbols.txt]
        - m.reference_path()/m.gene_synonyms_filename [default = v1/reference/gene_synonyms.txt]
        - m.dbxref_path()/*.txt [default = v1/reference/dbxrefs/*.txt]

    """

    raw_gene_info_filepath = os.path.join(m.raw_path(), m.gene_info_filename)
    reference_gene_symbols_filepath = os.path.join(
        m.reference_path(), m.gene_symbols_filename
    )
    reference_gene_synonyms_filepath = os.path.join(
        m.reference_path(), m.gene_synonyms_filename
    )
    dbxref_path = m.dbxref_path()
    verbose = m.verbose
    url = m.ncbi_gene_info_url

    # Download the gene symbols file
    download_gene_symbols(
        output_filepath=raw_gene_info_filepath, url=url, verbose=verbose
    )

    # Extract gene data
    gene_symbols, dbxrefs, gene_synonyms = extract_gene_data(
        filepath=raw_gene_info_filepath
    )

    # Save gene symbols to a file
    with open(reference_gene_symbols_filepath, "w") as file:
        for symbol in gene_symbols:
            file.write(symbol + "\n")

    msg2(verbose, f"Gene symbols saved to {reference_gene_symbols_filepath}")

    # Save dbXrefs to separate files
    for identifier in dbxrefs:
        identifier_parts = identifier.split(":")
        identifier_type = identifier_parts[0].replace("/", "_")
        identifier_value = ":".join(identifier_parts[1:])
        filename = f"{dbxref_path}/{identifier_type}.txt"
        with open(filename, "a") as file:
            file.write(identifier_value + "\n")

    """
    for identifiers in dbxrefs:
        for identifier in identifiers:
            identifier_parts = identifier.split(":")
            identifier_type = identifier_parts[0].replace('/','_')
            identifier_value = ":".join(identifier_parts[1:])
            filename = f"{dbxref_path}/{identifier_type}.txt"
            with open(filename, "a") as file:
                file.write(identifier_value + "\n")
    """
    msg2(verbose, "dbXrefs saved to individual files.")

    # Save gene synonyms to a file
    with open(reference_gene_synonyms_filepath, "w") as file:
        for synonym in gene_synonyms:
            file.write(synonym + "\n")

    msg2(verbose, f"Gene synonyms saved to {reference_gene_synonyms_filepath}")


def create_frequency_list(m: ReferenceData) -> List:
    """Creates a file that lists the most common English words, up to (m.corpus_stop_word_list_length) in size,
    and saves to m.search_path()/m.frequency_list_filename [default = v1/data/search_terms/frequency_list.txt]

    Returns:

      A list of most common words, most common first, all lower case

    """
    frequency_list_outpath = os.path.join(m.search_path(), m.frequency_list_filename)
    stop_word_list_length = m.corpus_stop_word_list_length
    verbose = m.verbose

    from nltk import FreqDist
    import string

    words = fetch_brown_corpus()

    # Remove punctuation and convert to lowercase
    words = [
        word.lower()
        for word in words
        if word not in string.punctuation and word.isalnum()
    ]

    # Compute the frequency distribution of words
    freq_dist = FreqDist(words)

    # Get the most frequent words
    most_common_words = freq_dist.most_common(stop_word_list_length)

    # Write the frequency list to a file
    with open(frequency_list_outpath, "w") as file:
        for word, frequency in most_common_words:
            file.write(word + "\n")
    msg2(verbose, f"Wrote {frequency_list_outpath}")
    return most_common_words


def create_search_terms_file(m: ReferenceData):
    """Uses the files from `create_gene_reference_data` to generate the list of terms for searching each abstract.

    Each abstract is searched for gene symbols, synonyms, and
    dbxref's. The file created here lists all the gene info terms to
    use for that search. The file is saved to
    m.search_path()/m.search_terms_filename [default =
    `v1/data/search_terms/search_terms.txt`].

    The file f"{m.search_path()/m.search_terms_filename}.unsorted" is a byproduct and can be discarded.

    Returns:

      A list of the search terms (gene symbols, synonyms and accession IDs)

    """

    import os

    dbxrefs = m.dbxrefs
    dbxrefs_path = m.dbxref_path()
    gene_symbols_filepath = os.path.join(m.reference_path(), m.gene_symbols_filename)
    gene_synonyms_filepath = os.path.join(m.reference_path(), m.gene_synonyms_filename)
    search_terms_filepath = os.path.join(m.search_path(), m.search_terms_filename)
    verbose = m.verbose

    if dbxrefs == []:
        # Get a list of all files in the directory
        dbxrefs = os.listdir(dbxrefs_path)
        # Filter out directories from the list
        dbxrefs = [f for f in dbxrefs if os.path.isfile(os.path.join(dbxrefs_path, f))]

    with open(f"{search_terms_filepath}.unsorted", "w") as outfile:
        for ref in dbxrefs:
            with open(os.path.join(dbxrefs_path, ref)) as infile:
                sorted_lines = sorted(set(infile.readlines()))
                outfile.writelines(sorted_lines)
                # outfile.write(infile.read())

        with open(gene_symbols_filepath) as infile:
            sorted_lines = sorted(set(infile.readlines()))
            outfile.writelines(sorted_lines)
            # outfile.write(infile.read())

        with open(gene_synonyms_filepath) as infile:
            sorted_lines = sorted(set(infile.readlines()))
            outfile.writelines(sorted_lines)
            # outfile.write(infile.read())

    # Sort and remove duplicates from the search terms file
    search_terms_unsorted_filepath = f"{search_terms_filepath}.unsorted"
    os.system(
        f"sort -u {search_terms_unsorted_filepath} | grep -v not > {search_terms_filepath}"
    )

    msg2(verbose, f"Created {search_terms_filepath}.")
    msg2(verbose, f"Created {search_terms_unsorted_filepath} - can be removed.")
    line_count = sum(1 for line in open(search_terms_filepath))
    msg2(verbose, f"Number of lines in {search_terms_filepath}: {line_count}")


def create_filtered_search_terms(m: ReferenceData) -> List:
    """Filters search_terms against m.frequency_list_filename and m.custom_stop_words.

    If the search term (e.g., gene name) is also a stop word (e.g.,
    'DANGER'), then retain the term's original case. Otherwise,
    lowercase the term. Either way, add it to the list of terms to be returned.

    Ultimately, when an abstract is searched for a filtered_term, it
    will only match case if the term is a stop_word. This allows for
    more permissive matching against gene names in abstracts that use
    peculiar case.

    Returns:

      List of the terms, all lowercase, accept for those that matched an English language word in the frequency_list.

    """

    search_file = os.path.join(m.search_path(), m.search_terms_filename)
    frequency_list_outpath = os.path.join(m.search_path(), m.frequency_list_filename)
    custom_words = m.custom_stop_words
    final_file = os.path.join(m.search_path(), m.filtered_terms_filename)
    verbose = m.verbose

    stop_words = create_stop_words(frequency_list_outpath, custom_words)

    search_terms = read_search_stop(search_file=search_file, stop_words=stop_words)
    msg2(verbose, f"Number of original search_terms:{len(search_terms)}")
    final_terms, filtered_terms, matched_stop_words = filter_search_terms(
        search_terms, stop_words
    )
    msg2(
        verbose,
        f"number of filtered_terms:{len(filtered_terms)}\nfinal number of final_terms:{len(final_terms)}\n number of matched_stop_words:{len(matched_stop_words)}\nmatched_stop_words={matched_stop_words}",
    )
    if final_file is not None:
        with open(final_file, "w") as f:
            f.writelines("\n".join(final_terms))
    msg2(verbose, f"Created {final_file}")
    return final_terms


def fetch_abstracts(m: ReferenceData, get_updates: bool = False):  # noqa: C901
    """Downloads files from m.ncbi_ftp_host.

    ARGS:

      get_updates: If True then download the updatefiles, otherwise get the baseline. See : https://ftp.ncbi.nlm.nih.gov/pubmed/baseline/README.txt

    """
    import subprocess

    num_files = m.num_abstract_xml_files
    refresh = m.refresh_abstract_xml_files
    ftp_host = m.ncbi_ftp_host
    download_dir = m.pub_inpath()
    verbose = m.verbose
    if get_updates:
        download_dir = os.path.join(download_dir, m.abstract_updatefiles_inpath)
        ftp_path = m.ncbi_ftp_updatefiles_path
        msg2(verbose, f"! Getting update files, download_dir={download_dir}.")
    else:
        ftp_path = m.ncbi_ftp_baseline_path

    """ This can probably be done faster with download_files.sh """
    msg2(verbose, f"Download Directory: {download_dir}")
    msg2(verbose, f"Number of abstracts to ensure have been downloaded: {num_files}")
    msg2(verbose, f"Refresh: {refresh}")

    # Retrieve file names and find the largest number
    # file_list = subprocess.check_output(['curl', '-s', f"ftp://{ftp_host}{ftp_path}"]).decode().splitlines()

    output = subprocess.check_output(
        ["curl", "-s", f"ftp://{ftp_host}{ftp_path}"]
    ).decode()
    msg2(verbose, f"file_list curl: curl -s ftp://{ftp_host}{ftp_path}")
    file_list = [
        line.split()[-1] for line in output.splitlines() if line.endswith(".xml.gz")
    ]

    msg2(verbose, f"Total number of NCBI abstract XML files: {len(file_list)}")
    latest_files = [
        file_name
        for file_name in file_list
        if file_name.startswith("pubmed23n") and file_name.endswith(".xml.gz")
    ]
    latest_files.sort(reverse=True)
    latest_files = latest_files[:num_files]
    msg2(verbose, f"latest_files {num_files}: {latest_files}")

    # Check if enough files are available
    import sys

    if len(latest_files) == 0:
        msg1(verbose, "Error: Insufficient number of files available!")
        sys.exit(1)

    # Calculate total predicted size
    total_size = 0
    for file_name in latest_files:
        response = subprocess.check_output(
            ["curl", "-sI", f"ftp://{ftp_host}{ftp_path}{file_name}"]
        ).decode()
        file_size = int(response.split("Content-Length: ")[1].split("\r")[0])
        total_size += file_size

    # Check disk space before downloading
    check_disk_space(total_size, download_dir, verbose=verbose)

    # Download and check files
    for file_name in latest_files:
        md5_file_name = f"{file_name}.md5"
        file_path = os.path.join(download_dir, file_name)
        md5_file_path = os.path.join(download_dir, md5_file_name)

        # Refresh files that were previously downloaded?
        if not refresh:
            # No, so skip downloading those again

            # If one file or the other is missing, you still have to do a download
            # Here, just provide information as to which files are present.
            if os.path.isfile(file_path) and not os.path.isfile(md5_file_path):
                msg1(verbose, f"ERROR: Missing - {md5_file_path}; re-downloading now")
            if not os.path.isfile(file_path) and os.path.isfile(md5_file_path):
                msg1(verbose, f"ERROR: Missing - {file_path}; re-downloading now")

            if os.path.isfile(file_path) and os.path.isfile(md5_file_path):
                msg1(verbose, f"SKIP: {file_path} exists.")
                continue

        # Check file size
        response = subprocess.check_output(
            ["curl", "-sI", f"ftp://{ftp_host}{ftp_path}{file_name}"]
        ).decode()
        file_size = int(response.split("Content-Length: ")[1].split("\r")[0])

        msg2(verbose, f"File: {file_name}, Size: {file_size} bytes")

        # Download file
        msg2(verbose, f"WARNING: Downloading: {file_name} to {download_dir}")
        if os.path.isfile(file_path):
            os.remove(file_path)
        download_file(f"ftp://{ftp_host}{ftp_path}{file_name}", file_path, verbose)

        # Download MD5 file
        if os.path.isfile(md5_file_path):
            os.remove(md5_file_path)
        download_file(
            f"ftp://{ftp_host}{ftp_path}{md5_file_name}", md5_file_path, verbose
        )

        # Check MD5
        verify_md5(file_path, md5_file_path, verbose)

    total_size_human = (
        subprocess.check_output(["numfmt", "--to=iec-i", "--suffix=B", str(total_size)])
        .decode()
        .strip()
    )
    msg2(verbose, f"Total size of abstract files: {total_size_human}")


def create_pubcsv_dataset(m: ReferenceData, parse_updates: bool = False) -> List:
    """Parse all the XML files into csv files. Overwites anything previously written. Takes about 14min for 30 (2-3 per minute)

    ARGS:

      parse_updates: If True then parse the updatefiles, otherwise parse the baseline. See : https://ftp.ncbi.nlm.nih.gov/pubmed/baseline/README.txt

    """
    import os

    abstract_length_threshold = m.abstract_length_threshold
    pub_inpath = m.pub_inpath()
    pub_updatefiles_inpath = os.path.join(m.pub_inpath(), m.abstract_updatefiles_inpath)
    if parse_updates:
        pub_inpath = os.path.join(pub_inpath, m.abstract_updatefiles_inpath)
    pub_outpath = m.pub_outpath()
    verbose = m.verbose

    csv_list = []

    # Iterate through files in the directory
    csv_list = parse_pubs(
        inpath=pub_inpath,
        outpath=pub_outpath,
        length_threshold=abstract_length_threshold,
        csv_list=csv_list,
        verbose=verbose,
    )

    # Repeat for 'updatefiles'
    csv_list = parse_pubs(
        inpath=pub_updatefiles_inpath,
        outpath=pub_outpath,
        length_threshold=abstract_length_threshold,
        csv_list=csv_list,
        verbose=verbose,
    )

    return csv_list


def create_gene_files(m: ReferenceData):
    """Calls the search.awk script in gpubs/scripts. This results in
    all the abstracts files being tagged with the gene
    name/identifier, if present. Uses the m.filtered_terms_filename for search terms.

    If the search term is also a 'stop word', then search.awk matches
    case from filtered_terms.txt (which ultimately comes from NCBI's
    gene_info.gz). Otherwise, case is permissive to accommodate
    abstracts that use peculiar case for a gene. Note that terms in
    filtered_terms.txt are also all lowercase for performance, accept
    those that are stop words.

    Note - to use a custom list of gene symbols/synonyms/accession
    ids, just filter everything out of the filtered_terms.txt except
    the genes of interest. Find the filtered_terms.txt file in
    m.search_path()/m.filtered_terms_filename [default:
    v1/data/search_terms/filtered_terms.txt]

    This is the last setp of the pipeline

    Exmample usage:

    >>> m = ReferenceData()
    >>> create_gene_reference_data(m) # get gene_info from NCBI
    >>> _ = create_frequency_list(m)  # create English stop words from 'brown' corpus
    >>> create_search_terms_file(m)   # combine the files parsed out of gene_info to get search terms
    >>> _ = create_filtered_search_terms(m)  # filter search terms with stop words
    >>> fetch_abstracts(m)            # download abstract XML files from NCBI
    >>> fetch_abstracts(m, get_updates=True) # download abstract XML "updatefiles" from NCBI
    >>> _ = create_pubcsv_dataset(m)  # translate abstract XML files to csv
    >>> # NOTE: keep only genes of interest in v1/data/search_terms/filtered_terms.txt
    >>> create_gene_files(m)          # tag abstracts with the search terms (genes) from filtered_terms.txt
    >>> # final files in m.genes_outpath [default= v1/data/csvpubs/genes]

    """

    filtered_terms_file = os.path.join(m.search_path(), m.filtered_terms_filename)
    csv_inpath = m.pub_outpath()
    csv_outpath = m.genes_outpath()
    verbose = m.verbose

    import glob
    import subprocess

    awk_script = "search.awk"
    # Check if awk_script is under ./scripts - e.g., this is being run from the notebook from inside the repo
    # otherwise awk_script should be in path - e.g. gpubs has been pip install'd
    if os.path.isfile(os.path.join("scripts", awk_script)):
        awk_script = os.path.join("scripts", awk_script)
    for file_name_path in glob.glob(os.path.join(csv_inpath, "pubmed*.xml.gz.csv")):
        file_name = os.path.basename(file_name_path)
        input_csv_file = os.path.join(csv_inpath, file_name)
        output_csv_file = os.path.join(csv_outpath, file_name)
        msg2(verbose, f"Creating {output_csv_file}")
        error_file = os.path.join(csv_outpath, f"{file_name}.err")
        # xxx test this out, then run make to install version 2
        command = [awk_script, filtered_terms_file, input_csv_file]
        with open(output_csv_file, "w") as output, open(error_file, "w") as error:
            subprocess.run(command, stdout=output, stderr=error)


def pipeline(m: ReferenceData):
    """Run the whole data pipeline, end to end. See QuickStart notebook for step-by-step outputs

    Example usage:

      >>> import gpubs
      >>> from gpubs.models import ReferenceData
      >>> from gpubs.api import pipeline
      >>>
      >>> # Create data model
      >>> # exclude miRNA and MIM dbxrefs
      >>> m = ReferenceData(version = "../../v1",       # put data above the github repo
      >>>                   verbose = 2,                # print info messages
      >>>                   num_abstract_xml_files = 3, # only download 3 abstract XML files from NCBI
      >>>                   dbxrefs = ["AllianceGenome.txt", "Ensembl.txt", "HGNC.txt", "IMGT_GENE-DB.txt"])
      >>> pipeline(m)

    """

    # Fetch data/raw/gene_info.gz
    # and create the human genes lists under data/reference (gene_symbols.txt, gene_synonyms.txt, dbxrefs/*)
    create_gene_reference_data(m)

    # The goal of the following 3 calls is to
    # create data/search_terms/filtered_terms.txt from english language corpus

    # Create a word frequency list from an English language corpus
    _ = create_frequency_list(m)

    # Create the file of gene search terms (data/search_terms/search_terms.txt) using stop words from frequency list
    create_search_terms_file(m)

    # Create the filtered_terms.txt file
    _ = create_filtered_search_terms(m)

    # Fetch NCBI article zips
    # - There are 1165 files (2023) with about 15000 abstracts each.
    # - ~60GB is needed to get all files
    # - On AWS you can get them all in under an hour. 5 hours on consumer-grade wifi.
    fetch_abstracts(m)

    # Get NCBI
    # ~ 222 abstracts in June 2023
    fetch_abstracts(m, get_updates=True)

    # Create CSVs from XMLs
    # - This takes about 3 minutes to do 10 files; or about 5 hours to do them all
    _ = create_pubcsv_dataset(m)

    # Create new CSVs that include GENES column under data/csvpubs/genes
    # - Takes about 40s for 10 files, which is much slower than just running the awk script
    # - With default settings, it filters out about 42% of the abstracts, most of which are 2022
    create_gene_files(m)



# replace these everywhere
def msginfo(verbose, msg):
    msg2(verbose, msg)


def msg2(verbose, msg):
    if verbose >= 2:
        print(msg)


def msg1(verbose, msg):
    if verbose >= 1:
        print(msg)



import os
from typing import List, Dict
from pydantic import BaseModel, root_validator

from gpubs.log import msg2


class ReferenceData(BaseModel):
    """
    Files for retrieving and transforming reference gene information
    """

    def __init__(self, *args, **kwargs):
        super(ReferenceData, self).__init__(*args, **kwargs)

        self.version_root = os.path.join(os.getcwd(), str(self.version), self.data_root)
        msg2(self.verbose, f"version_root={self.version_root}")

        # make directory structure
        os.makedirs(self.version_root, exist_ok=True)
        os.makedirs(self.raw_path(), exist_ok=True)
        os.makedirs(self.reference_path(), exist_ok=True)
        os.makedirs(self.dbxref_path(), exist_ok=True)
        os.makedirs(self.search_path(), exist_ok=True)
        os.makedirs(self.pub_inpath(), exist_ok=True)
        os.makedirs(
            os.path.join(self.pub_inpath(), self.abstract_updatefiles_inpath),
            exist_ok=True,
        )
        os.makedirs(self.pub_outpath(), exist_ok=True)
        os.makedirs(self.genes_outpath(), exist_ok=True)

        msg2(self.verbose, "Created directory structure.")

    ncbi_gene_info_url: str = "https://ftp.ncbi.nlm.nih.gov/gene/DATA/GENE_INFO/Mammalia/Homo_sapiens.gene_info.gz"

    ncbi_ftp_host: str = "ftp.ncbi.nlm.nih.gov"

    ncbi_ftp_baseline_path: str = "/pubmed/baseline/"

    ncbi_ftp_updatefiles_path: str = "/pubmed/updatefiles/"

    data_root: str = "data/"
    """ Full path to where all the raw, reference, and generated data are stored """

    raw_data_path: str = "raw/"

    reference_data_path: str = "reference/"

    dbxref_reference_data_path: str = "dbxrefs/"

    dbxrefs: List = []
    """ If empty, it will just get filled with a list of all the files created in the m.dbxref_path() """

    gene_info_filename: str = "gene_info.gz"

    gene_symbols_filename: str = "gene_symbols.txt"

    gene_synonyms_filename: str = "gene_synonyms.txt"

    search_terms_path: str = "search_terms/"

    frequency_list_filename: str = "frequency_list.txt"

    corpus_stop_word_list_length: int = 4000

    # may overlap somewhat with stop words
    custom_stop_words: List = [
        "ago",
        "aim",
        "amid",
        "april",
        "arch",
        "bed",
        "bite",
        "bug",
        "cage",
        "co",
        "crop",
        "damage",
        "danger",
        "digit",
        "et",
        "fast",
        "fat",
        "fate",
        "fire",
        "flower",
        "gap",
        "genesis",
        "gov",
        "gpa",
        "grasp",
        "ii",
        "inos",
        "iv",
        "killer",
        "lab",
        "lamp",
        "laser",
        "map",
        "mask",
        "mater",
        "melt",
        "mice",
        "minor",
        "miss",
        "mv",
        "nail",
        "net",
        "not",
        "osf",
        "pan",
        "par",
        "pha",
        "rab",
        "race",
        "rain",
        "rank",
        "san",
        "sand",
        "se",
        "sink",
        "soft",
        "spatial",
        "spin",
        "spp",
        "steel",
        "stop",
        "storm",
        "tactile",
        "tau",
        "theta",
        "tip",
        "traits",
        "via",
        "apex",
        "app",
        "apps",
        "args",
        "ash",
        "bar",
        "bit",
        "cast",
        "cats",
        "cava",
        "cd",
        "clock",
        "coil",
        "cope",
        "cord",
        "delta",
        "eat",
        "eg",
        "fats",
        "fish",
        "fix",
        "flame",
        "flap",
        "fuse",
        "grid",
        "gum",
        "heal",
        "hip",
        "hits",
        "hub",
        "igm",
        "li",
        "maps",
        "mets",
        "mix",
        "mn",
        "ms",
        "nm",
        "nodal",
        "pigs",
        "prey",
        "pros",
        "pt",
        "pva",
        "ray",
        "sac",
        "scar",
        "sea",
        "sea",
        "sp",
        "steep",
        "tank",
        "tied",
        "toll",
        "trap",
        "wire",
    ]

    search_terms_filename: str = "search_terms.txt"

    filtered_terms_filename: str = "filtered_terms.txt"

    abstract_inpath: str = "pubs/"

    abstract_updatefiles_inpath: str = "updates/"
    """ This path gets added on to abstract_inpath (or pub_inpath) """

    refresh_abstract_xml_files: bool = False
    """ Set to True to overwrite downloaded NCBI XML abstract files and checksum files """

    num_abstract_xml_files: int
    """ Number of NCBI XML files to download; Set this to -1 to get all abstracts """

    abstract_outpath: str = "csvpubs/"

    abstract_length_threshold: int = 405

    abstract_genes_outpath: str = "genes/"

    # path functions
    def raw_path(self):
        return os.path.join(os.getcwd(), self.version_root, self.raw_data_path)

    def reference_path(self):
        return os.path.join(os.getcwd(), self.version_root, self.reference_data_path)

    def dbxref_path(self):
        return os.path.join(
            os.getcwd(),
            self.version_root,
            self.reference_data_path,
            self.dbxref_reference_data_path,
        )

    def search_path(self):
        return os.path.join(os.getcwd(), self.version_root, self.search_terms_path)

    def pub_inpath(self):
        return os.path.join(
            os.getcwd(), self.version_root, self.raw_data_path, self.abstract_inpath
        )

    def pub_outpath(self):
        return os.path.join(os.getcwd(), self.version_root, self.abstract_outpath)

    def genes_outpath(self):
        return os.path.join(self.pub_outpath(), self.abstract_genes_outpath)

    verbose: int = 0
    """ 0 prints nothing, 1 prints errors and warnings, 2 prints info """

    debug: bool = False
    """ Prints very detailed debuggin messages """

    version: str = None
    """ If None, version will e set to timestamp """

    rand_seed: int = 42
    """ If None, reproducible within the same release"""

    version_root: str = None
    """ Leave this alone it will be computed as data_root/version """

    @root_validator(pre=False, skip_on_failure=True)
    def valid_paths_for_data_tree(cls, v: Dict) -> Dict:
        data_root = v.get("data_root")
        version = v.get("version")

        import os

        if version is None:
            import subprocess

            # use miliseconds and create new directory structure
            version = str(
                subprocess.check_output(["date", "+%1N.%S_%M_%H_%Y_%m_%d"])
                .strip()
                .decode()
            )
            v["version"] = version

        if data_root is None:
            raise Exception("Error: data_root cannot be None.")

        version_root = os.path.join(os.getcwd(), data_root, str(version))
        v["version_root"] = version_root

        if os.path.exists(version_root):
            msg2(f"Using existing data director {version_root}")

        return v



import os
import subprocess
import urllib.request
from urllib.error import URLError
from gpubs.log import msg1, msg2


def check_disk_space(predicted_size, download_dir, verbose):
    required_space = predicted_size
    available_space = (
        os.statvfs(download_dir).f_frsize * os.statvfs(download_dir).f_bavail
    )
    required_space_human = (
        subprocess.check_output(
            ["numfmt", "--to=iec-i", "--suffix=B", str(required_space)]
        )
        .decode()
        .strip()
    )
    available_space_human = (
        subprocess.check_output(
            ["numfmt", "--to=iec-i", "--suffix=B", str(available_space)]
        )
        .decode()
        .strip()
    )

    msg1(
        verbose,
        f"Predicted download size = {required_space_human}, Available space = {available_space_human}",
    )

    if required_space > available_space:
        print(
            f"Insufficient disk space! Required: {required_space_human}, Available: {available_space_human}"
        )
        exit(1)


def download_file(url, file_path, verbose):
    try:
        urllib.request.urlretrieve(url, file_path)
    except URLError as e:
        msg1(verbose, f"Error downloading file: {url}")
        msg1(verbose, f"Reason: {str(e.reason)}")
        # exit(1)


def verify_md5(file_path, md5_file_path, verbose):
    try:
        output = subprocess.check_output(
            ["md5sum", "-c", os.path.basename(md5_file_path)],
            cwd=os.path.dirname(md5_file_path),
            stderr=subprocess.DEVNULL,
        ).decode()
        # output = subprocess.check_output(['md5sum', '-c', md5_file_path], stderr=subprocess.DEVNULL).decode()
        if "OK" in output:
            msg2(verbose, f"{md5_file_path}: OK - MD5 checksum verification succeeded.")
        else:
            msg1(
                verbose,
                f"ERROR: {md5_file_path}: FAILED - MD5 checksum verification failed.",
            )
    except subprocess.CalledProcessError:
        msg1(
            verbose,
            f"ERROR: {md5_file_path}: FAILED - MD5 checksum verification failed.",
        )



from typing import Set
import nltk


def fetch_brown_corpus():
    from nltk.corpus import brown

    corpus = "brown"
    # Select a specific category from the Brown Corpus for analysis
    category = "learned"
    # Load the Brown Corpus
    nltk.download(corpus)
    # Get the words from the selected category
    words = brown.words(categories=category)
    return words


def read_search_stop(search_file, stop_words):
    # Load search terms from the file into a set
    search_terms = []
    with open(search_file, "r") as f:
        search_terms = [line.strip() for line in f]
    return search_terms


def filter_search_terms(search_terms, stop_words):
    filtered_terms = set()  # use a set to avoid duplicates
    matched_stop_words = []

    for term in search_terms:
        if term.lower() in stop_words:
            # keep stop_words in original case
            matched_stop_words.append(term)
        else:
            # lower case non-stop_words for permissive matching
            filtered_terms.add(term.lower())

    # add the two lists together so the matched terms retain original case
    # this will allow for finding genes that are also common english words
    final_terms = list(filtered_terms) + matched_stop_words
    return (final_terms, filtered_terms, matched_stop_words)


def create_stop_words(frequency_list_outpath, custom_words) -> Set:
    from nltk.corpus import stopwords

    # Download the stopwords corpus
    nltk.download("stopwords")

    # Load the existing stop words
    stop_words = set(stopwords.words("english"))

    # Read the words from the frequency_list.txt file
    with open(frequency_list_outpath, "r") as file:
        frequency_words = file.read().splitlines()

    # Add the frequency words to the stop words set
    stop_words.update(frequency_words)

    # Add custom words
    stop_words.update(custom_words)
    return stop_words



from gpubs.log import msg2
import requests


def download_gene_symbols(
    output_filepath="data/raw/gene_info.gz",
    url="https://ftp.ncbi.nlm.nih.gov/gene/DATA/GENE_INFO/Mammalia/Homo_sapiens.gene_info.gz",
    verbose=0,
):
    response = requests.get(url)
    response.raise_for_status()

    with open(output_filepath, "wb") as file:
        file.write(response.content)

    msg2(verbose, "Download completed.")


def extract_gene_data(filepath="data/raw/gene_info.gz"):
    """returns 3 sets"""
    import gzip

    gene_symbols = set()
    dbxrefs = set()
    gene_synonyms = set()

    with gzip.open(filepath, "rt") as file:
        for line in file:
            if not line.startswith("#"):
                fields = line.strip().split("\t")
                gene_symbols.add(fields[2])

                # Extract dbXrefs
                dbxref_field = fields[5]
                if dbxref_field != "-":
                    identifiers = dbxref_field.split("|")
                    dbxrefs.update(identifiers)

                # Extract gene synonyms
                synonym_field = fields[4]
                if synonym_field != "-":
                    synonyms = synonym_field.split("|")
                    gene_synonyms.update(synonyms)

    return gene_symbols, dbxrefs, gene_synonyms



from setuptools import setup  # noga: F401

import json

with open("./release-info.json") as f:
    j = json.load(f)
    # overwrite the version file from source control so the library will
    # have the latest version number created from release-info.json
    with open("./gpubs/_version.py", "w") as f:
        f.write(f'__version__ = "{j["VERSION"]}"\n')
        f.flush()

        setup(
            name=j["LIBNAME"],
            packages=[j["LIBNAME"]],
            description=j["DESCRIPTION"],
            version=f"{j['VERSION']}",
            url=f'http://github.org/{j["REPONAME"]}',
            author=j["AUTHOR"],
            author_email=j["EMAIL"],
            keywords=j["KEYWORDS"],
            include_package_data=True,
            install_requires=[
                "ipykernel==6.19.2",
                "lxml==4.9.2",
                "nltk==3.7",
                "pandas==1.5.3",
                "pydantic==1.10.8",
                "requests==2.29.0",
            ],
            scripts=[
                "scripts/create_search_terms_file.sh",
                "scripts/search.awk",
                "scripts/download_pubs.sh",
            ],
        )
